package org.eclipse.jdt.internal.compiler.lookup;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.MethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.ReferenceContext;
import org.eclipse.jdt.internal.compiler.lookup.BinaryTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.ImplicitNullAnnotationVerifier;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.PackageBinding;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedGenericMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.Substitution;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.util.HashtableOfObject;
import org.eclipse.jdt.internal.compiler.util.SimpleSet;
import org.eclipse.jdt.internal.compiler.util.Sorting;

public abstract class MethodVerifier extends ImplicitNullAnnotationVerifier {
  SourceTypeBinding type;
  
  HashtableOfObject inheritedMethods;
  
  HashtableOfObject currentMethods;
  
  HashtableOfObject inheritedOverriddenMethods;
  
  MethodVerifier(LookupEnvironment environment) {
    super(environment);
    this.type = null;
    this.inheritedMethods = null;
    this.currentMethods = null;
    this.inheritedOverriddenMethods = null;
  }
  
  boolean areMethodsCompatible(MethodBinding one, MethodBinding two) {
    return areMethodsCompatible(one, two, this.environment);
  }
  
  static boolean areMethodsCompatible(MethodBinding one, MethodBinding two, LookupEnvironment environment) {
    one = one.original();
    two = one.findOriginalInheritedMethod(two);
    if (two == null)
      return false; 
    return isParameterSubsignature(one, two, environment);
  }
  
  boolean areReturnTypesCompatible(MethodBinding one, MethodBinding two) {
    return areReturnTypesCompatible(one, two, this.type.scope.environment());
  }
  
  public static boolean areReturnTypesCompatible(MethodBinding one, MethodBinding two, LookupEnvironment environment) {
    if (TypeBinding.equalsEquals(one.returnType, two.returnType))
      return true; 
    if (environment.globalOptions.sourceLevel >= 3211264L) {
      if (one.returnType.isBaseType())
        return false; 
      if (!one.declaringClass.isInterface() && one.declaringClass.id == 1)
        return two.returnType.isCompatibleWith(one.returnType); 
      return one.returnType.isCompatibleWith(two.returnType);
    } 
    return areTypesEqual(one.returnType.erasure(), two.returnType.erasure());
  }
  
  boolean canSkipInheritedMethods() {
    if (this.type.superclass() != null && this.type.superclass().isAbstract())
      return false; 
    return (this.type.superInterfaces() == Binding.NO_SUPERINTERFACES);
  }
  
  boolean canSkipInheritedMethods(MethodBinding one, MethodBinding two) {
    return !(two != null && 
      !TypeBinding.equalsEquals((TypeBinding)one.declaringClass, (TypeBinding)two.declaringClass));
  }
  
  void checkAbstractMethod(MethodBinding abstractMethod) {
    if (mustImplementAbstractMethod(abstractMethod.declaringClass)) {
      TypeDeclaration typeDeclaration = this.type.scope.referenceContext;
      if (typeDeclaration != null) {
        MethodDeclaration missingAbstractMethod = typeDeclaration.addMissingAbstractMethodFor(abstractMethod);
        missingAbstractMethod.scope.problemReporter().abstractMethodMustBeImplemented(this.type, abstractMethod);
      } else {
        problemReporter().abstractMethodMustBeImplemented(this.type, abstractMethod);
      } 
    } 
  }
  
  void checkAgainstInheritedMethods(MethodBinding currentMethod, MethodBinding[] methods, int length, MethodBinding[] allInheritedMethods) {
    if (this.type.isAnnotationType()) {
      problemReporter().annotationCannotOverrideMethod(currentMethod, methods[length - 1]);
      return;
    } 
    CompilerOptions options = this.type.scope.compilerOptions();
    int[] overriddenInheritedMethods = (length > 1) ? findOverriddenInheritedMethods(methods, length) : null;
    int i;
    label78: for (i = length; --i >= 0; ) {
      MethodBinding inheritedMethod = methods[i];
      if (overriddenInheritedMethods == null || overriddenInheritedMethods[i] == 0) {
        if (currentMethod.isStatic() != inheritedMethod.isStatic()) {
          problemReporter(currentMethod).staticAndInstanceConflict(currentMethod, inheritedMethod);
          continue;
        } 
        if (inheritedMethod.isAbstract()) {
          if (inheritedMethod.declaringClass.isInterface()) {
            currentMethod.modifiers |= 0x20000000;
          } else {
            currentMethod.modifiers |= 0x30000000;
          } 
        } else if (inheritedMethod.isPublic() || !this.type.isInterface()) {
          if (currentMethod.isDefaultMethod() && 
            !inheritedMethod.isFinal() && 
            inheritedMethod.declaringClass.id == 1) {
            problemReporter(currentMethod).defaultMethodOverridesObjectMethod(currentMethod);
          } else if (inheritedMethod.isDefaultMethod()) {
            currentMethod.modifiers |= 0x20000000;
          } else {
            currentMethod.modifiers |= 0x10000000;
          } 
        } 
        if (!areReturnTypesCompatible(currentMethod, inheritedMethod) && (
          currentMethod.returnType.tagBits & 0x80L) == 0L && 
          reportIncompatibleReturnTypeError(currentMethod, inheritedMethod))
          continue; 
        reportRawReferences(currentMethod, inheritedMethod);
        if (currentMethod.thrownExceptions != Binding.NO_EXCEPTIONS)
          checkExceptions(currentMethod, inheritedMethod); 
        if (inheritedMethod.isFinal())
          problemReporter(currentMethod).finalMethodCannotBeOverridden(currentMethod, inheritedMethod); 
        if (!isAsVisible(currentMethod, inheritedMethod))
          problemReporter(currentMethod).visibilityConflict(currentMethod, inheritedMethod); 
        if (inheritedMethod.isSynchronized() && !currentMethod.isSynchronized())
          problemReporter(currentMethod).missingSynchronizedOnInheritedMethod(currentMethod, inheritedMethod); 
        if (options.reportDeprecationWhenOverridingDeprecatedMethod && inheritedMethod.isViewedAsDeprecated() && (
          !currentMethod.isViewedAsDeprecated() || options.reportDeprecationInsideDeprecatedCode)) {
          ReferenceBinding declaringClass = inheritedMethod.declaringClass;
          if (declaringClass.isInterface())
            for (int j = length; --j >= 0;) {
              if (i != j && (methods[j]).declaringClass.implementsInterface(declaringClass, false))
                continue label78; 
            }  
          problemReporter(currentMethod).overridesDeprecatedMethod(currentMethod, inheritedMethod);
        } 
      } 
      if (!inheritedMethod.isStatic() && !inheritedMethod.isFinal())
        checkForBridgeMethod(currentMethod, inheritedMethod, allInheritedMethods); 
    } 
    MethodBinding[] overridden = (MethodBinding[])this.inheritedOverriddenMethods.get(currentMethod.selector);
    if (overridden != null)
      for (int j = overridden.length; --j >= 0; ) {
        MethodBinding inheritedMethod = overridden[j];
        if (isParameterSubsignature(currentMethod, inheritedMethod) && 
          !inheritedMethod.isStatic() && !inheritedMethod.isFinal())
          checkForBridgeMethod(currentMethod, inheritedMethod, allInheritedMethods); 
      }  
  }
  
  void addBridgeMethodCandidate(MethodBinding overriddenMethod) {
    MethodBinding[] existing = (MethodBinding[])this.inheritedOverriddenMethods.get(overriddenMethod.selector);
    if (existing == null) {
      existing = new MethodBinding[] { overriddenMethod };
    } else {
      int length = existing.length;
      System.arraycopy(existing, 0, existing = new MethodBinding[length + 1], 0, length);
      existing[length] = overriddenMethod;
    } 
    this.inheritedOverriddenMethods.put(overriddenMethod.selector, existing);
  }
  
  public void reportRawReferences(MethodBinding currentMethod, MethodBinding inheritedMethod) {}
  
  void checkConcreteInheritedMethod(MethodBinding concreteMethod, MethodBinding[] abstractMethods) {
    if (concreteMethod.isStatic())
      problemReporter().staticInheritedMethodConflicts(this.type, concreteMethod, abstractMethods); 
    if (!concreteMethod.isPublic()) {
      int index = 0, length = abstractMethods.length;
      if (concreteMethod.isProtected()) {
        for (; index < length && 
          !abstractMethods[index].isPublic(); index++);
      } else if (concreteMethod.isDefault()) {
        for (; index < length && 
          abstractMethods[index].isDefault(); index++);
      } 
      if (index < length)
        problemReporter().inheritedMethodReducesVisibility(this.type, concreteMethod, abstractMethods); 
    } 
    if (concreteMethod.thrownExceptions != Binding.NO_EXCEPTIONS)
      for (int i = abstractMethods.length; --i >= 0;)
        checkExceptions(concreteMethod, abstractMethods[i]);  
    if (concreteMethod.isOrEnclosedByPrivateType())
      (concreteMethod.original()).modifiers |= 0x8000000; 
  }
  
  void checkExceptions(MethodBinding newMethod, MethodBinding inheritedMethod) {
    ReferenceBinding[] newExceptions = resolvedExceptionTypesFor(newMethod);
    ReferenceBinding[] inheritedExceptions = resolvedExceptionTypesFor(inheritedMethod);
    for (int i = newExceptions.length; --i >= 0; ) {
      ReferenceBinding newException = newExceptions[i];
      int j = inheritedExceptions.length;
      do {
      
      } while (--j > -1 && !isSameClassOrSubclassOf(newException, inheritedExceptions[j]));
      if (j == -1 && 
        !newException.isUncheckedException(false) && (
        newException.tagBits & 0x80L) == 0L)
        problemReporter(newMethod).incompatibleExceptionInThrowsClause(this.type, newMethod, inheritedMethod, newException); 
    } 
  }
  
  void checkForBridgeMethod(MethodBinding currentMethod, MethodBinding inheritedMethod, MethodBinding[] allInheritedMethods) {}
  
  void checkForMissingHashCodeMethod() {
    MethodBinding[] choices = this.type.getMethods(TypeConstants.EQUALS);
    boolean overridesEquals = false;
    for (int i = choices.length; !overridesEquals && --i >= 0;)
      overridesEquals = ((choices[i]).parameters.length == 1 && ((choices[i]).parameters[0]).id == 1); 
    if (overridesEquals) {
      MethodBinding hashCodeMethod = this.type.getExactMethod(TypeConstants.HASHCODE, Binding.NO_PARAMETERS, null);
      if (hashCodeMethod != null && hashCodeMethod.declaringClass.id == 1)
        problemReporter().shouldImplementHashcode(this.type); 
    } 
  }
  
  void checkForRedundantSuperinterfaces(ReferenceBinding superclass, ReferenceBinding[] superInterfaces) {
    if (superInterfaces == Binding.NO_SUPERINTERFACES)
      return; 
    SimpleSet interfacesToCheck = new SimpleSet(superInterfaces.length);
    SimpleSet redundantInterfaces = null;
    for (int i = 0, l = superInterfaces.length; i < l; i++) {
      ReferenceBinding toCheck = superInterfaces[i];
      for (int k = 0; k < l; k++) {
        ReferenceBinding implementedInterface = superInterfaces[k];
        if (i != k && toCheck.implementsInterface(implementedInterface, true)) {
          if (redundantInterfaces == null) {
            redundantInterfaces = new SimpleSet(3);
          } else if (redundantInterfaces.includes(implementedInterface)) {
            continue;
          } 
          redundantInterfaces.add(implementedInterface);
          TypeReference[] refs = this.type.scope.referenceContext.superInterfaces;
          for (int r = 0, rl = refs.length; r < rl; r++) {
            if (TypeBinding.equalsEquals((refs[r]).resolvedType, (TypeBinding)toCheck)) {
              problemReporter().redundantSuperInterface(this.type, refs[k], implementedInterface, toCheck);
              break;
            } 
          } 
        } 
        continue;
      } 
      interfacesToCheck.add(toCheck);
    } 
    ReferenceBinding[] itsInterfaces = null;
    SimpleSet inheritedInterfaces = new SimpleSet(5);
    ReferenceBinding superType = superclass;
    while (superType != null && superType.isValidBinding()) {
      if ((itsInterfaces = superType.superInterfaces()) != Binding.NO_SUPERINTERFACES)
        for (int k = 0, m = itsInterfaces.length; k < m; k++) {
          ReferenceBinding inheritedInterface = itsInterfaces[k];
          if (!inheritedInterfaces.includes(inheritedInterface) && inheritedInterface.isValidBinding()) {
            if (interfacesToCheck.includes(inheritedInterface)) {
              if (redundantInterfaces == null) {
                redundantInterfaces = new SimpleSet(3);
              } else if (redundantInterfaces.includes(inheritedInterface)) {
                continue;
              } 
              redundantInterfaces.add(inheritedInterface);
              TypeReference[] refs = this.type.scope.referenceContext.superInterfaces;
              for (int r = 0, rl = refs.length; r < rl; r++) {
                if (TypeBinding.equalsEquals((refs[r]).resolvedType, (TypeBinding)inheritedInterface)) {
                  problemReporter().redundantSuperInterface(this.type, refs[r], inheritedInterface, superType);
                  break;
                } 
              } 
              continue;
            } 
            inheritedInterfaces.add(inheritedInterface);
          } 
          continue;
        }  
      superType = superType.superclass();
    } 
    int nextPosition = inheritedInterfaces.elementSize;
    if (nextPosition == 0)
      return; 
    ReferenceBinding[] interfacesToVisit = new ReferenceBinding[nextPosition];
    inheritedInterfaces.asArray((Object[])interfacesToVisit);
    for (int j = 0; j < nextPosition; j++) {
      superType = interfacesToVisit[j];
      if ((itsInterfaces = superType.superInterfaces()) != Binding.NO_SUPERINTERFACES) {
        int itsLength = itsInterfaces.length;
        if (nextPosition + itsLength >= interfacesToVisit.length)
          System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition); 
        for (int a = 0; a < itsLength; a++) {
          ReferenceBinding inheritedInterface = itsInterfaces[a];
          if (!inheritedInterfaces.includes(inheritedInterface) && inheritedInterface.isValidBinding()) {
            if (interfacesToCheck.includes(inheritedInterface)) {
              if (redundantInterfaces == null) {
                redundantInterfaces = new SimpleSet(3);
              } else if (redundantInterfaces.includes(inheritedInterface)) {
                continue;
              } 
              redundantInterfaces.add(inheritedInterface);
              TypeReference[] refs = this.type.scope.referenceContext.superInterfaces;
              for (int r = 0, rl = refs.length; r < rl; r++) {
                if (TypeBinding.equalsEquals((refs[r]).resolvedType, (TypeBinding)inheritedInterface)) {
                  problemReporter().redundantSuperInterface(this.type, refs[r], inheritedInterface, superType);
                  break;
                } 
              } 
              continue;
            } 
            inheritedInterfaces.add(inheritedInterface);
            interfacesToVisit[nextPosition++] = inheritedInterface;
          } 
          continue;
        } 
      } 
    } 
  }
  
  void checkInheritedMethods(MethodBinding[] methods, int length, boolean[] isOverridden, boolean[] isInherited) {
    MethodBinding concreteMethod = (this.type.isInterface() || methods[0].isAbstract()) ? null : methods[0];
    if (concreteMethod == null) {
      MethodBinding bestAbstractMethod = (length == 1) ? methods[0] : findBestInheritedAbstractOrDefaultMethod(methods, length);
      boolean noMatch = (bestAbstractMethod == null);
      if (noMatch)
        bestAbstractMethod = methods[0]; 
      if (mustImplementAbstractMethod(bestAbstractMethod.declaringClass)) {
        TypeDeclaration typeDeclaration = this.type.scope.referenceContext;
        MethodBinding superclassAbstractMethod = methods[0];
        if (superclassAbstractMethod == bestAbstractMethod || superclassAbstractMethod.declaringClass.isInterface()) {
          if (typeDeclaration != null) {
            MethodDeclaration missingAbstractMethod = typeDeclaration.addMissingAbstractMethodFor(bestAbstractMethod);
            missingAbstractMethod.scope.problemReporter().abstractMethodMustBeImplemented(this.type, bestAbstractMethod);
          } else {
            problemReporter().abstractMethodMustBeImplemented(this.type, bestAbstractMethod);
          } 
        } else if (typeDeclaration != null) {
          MethodDeclaration missingAbstractMethod = typeDeclaration.addMissingAbstractMethodFor(bestAbstractMethod);
          missingAbstractMethod.scope.problemReporter().abstractMethodMustBeImplemented(this.type, bestAbstractMethod, superclassAbstractMethod);
        } else {
          problemReporter().abstractMethodMustBeImplemented(this.type, bestAbstractMethod, superclassAbstractMethod);
        } 
      } else if (noMatch) {
        problemReporter().inheritedMethodsHaveIncompatibleReturnTypes(this.type, methods, length, isOverridden);
      } 
      return;
    } 
    if (length < 2)
      return; 
    int index = length;
    do {
    
    } while (--index > 0 && checkInheritedReturnTypes(concreteMethod, methods[index]));
    if (index > 0) {
      MethodBinding bestAbstractMethod = findBestInheritedAbstractOrDefaultMethod(methods, length);
      if (bestAbstractMethod == null) {
        problemReporter().inheritedMethodsHaveIncompatibleReturnTypes(this.type, methods, length, isOverridden);
      } else {
        problemReporter().abstractMethodMustBeImplemented(this.type, bestAbstractMethod, concreteMethod);
      } 
      return;
    } 
    MethodBinding[] abstractMethods = new MethodBinding[length - 1];
    index = 0;
    for (int i = 0; i < length; i++) {
      if (methods[i].isAbstract() || (methods[i] != concreteMethod && methods[i].isDefaultMethod()))
        abstractMethods[index++] = methods[i]; 
    } 
    if (index == 0)
      return; 
    if (index < abstractMethods.length)
      System.arraycopy(abstractMethods, 0, abstractMethods = new MethodBinding[index], 0, index); 
    checkConcreteInheritedMethod(concreteMethod, abstractMethods);
  }
  
  boolean checkInheritedReturnTypes(MethodBinding method, MethodBinding otherMethod) {
    if (areReturnTypesCompatible(method, otherMethod))
      return true; 
    if (!this.type.isInterface() && (
      method.declaringClass.isClass() || !this.type.implementsInterface(method.declaringClass, false)) && (
      otherMethod.declaringClass.isClass() || !this.type.implementsInterface(otherMethod.declaringClass, false)))
      return true; 
    return false;
  }
  
  abstract void checkMethods();
  
  void checkPackagePrivateAbstractMethod(MethodBinding abstractMethod) {
    PackageBinding necessaryPackage = abstractMethod.declaringClass.fPackage;
    if (necessaryPackage == this.type.fPackage)
      return; 
    ReferenceBinding superType = this.type.superclass();
    char[] selector = abstractMethod.selector;
    do {
      if (!superType.isValidBinding())
        return; 
      if (!superType.isAbstract())
        return; 
      if (necessaryPackage != superType.fPackage)
        continue; 
      MethodBinding[] methods = superType.getMethods(selector);
      for (int m = methods.length; --m >= 0; ) {
        MethodBinding method = methods[m];
        if (method.isPrivate() || method.isConstructor() || method.isDefaultAbstract())
          continue; 
        if (areMethodsCompatible(method, abstractMethod))
          return; 
      } 
    } while (TypeBinding.notEquals((TypeBinding)(superType = superType.superclass()), (TypeBinding)abstractMethod.declaringClass));
    problemReporter().abstractMethodCannotBeOverridden(this.type, abstractMethod);
  }
  
  void computeInheritedMethods() {
    ReferenceBinding superclass = this.type.isInterface() ? 
      this.type.scope.getJavaLangObject() : 
      this.type.superclass();
    computeInheritedMethods(superclass, this.type.superInterfaces());
    checkForRedundantSuperinterfaces(superclass, this.type.superInterfaces());
  }
  
  void computeInheritedMethods(ReferenceBinding superclass, ReferenceBinding[] superInterfaces) {
    this.inheritedMethods = new HashtableOfObject(51);
    this.inheritedOverriddenMethods = new HashtableOfObject(11);
    ReferenceBinding superType = superclass;
    HashtableOfObject nonVisibleDefaultMethods = new HashtableOfObject(3);
    while (superType != null && superType.isValidBinding()) {
      MethodBinding[] methods = superType.unResolvedMethods();
      for (int m = methods.length; --m >= 0; ) {
        MethodBinding inheritedMethod = methods[m];
        if (inheritedMethod.isPrivate() || inheritedMethod.isConstructor() || inheritedMethod.isDefaultAbstract())
          continue; 
        MethodBinding[] existingMethods = (MethodBinding[])this.inheritedMethods.get(inheritedMethod.selector);
        if (existingMethods != null)
          for (int j = 0, length = existingMethods.length; j < length; j++) {
            MethodBinding existingMethod = existingMethods[j];
            if (TypeBinding.notEquals((TypeBinding)existingMethod.declaringClass, (TypeBinding)inheritedMethod.declaringClass) && areMethodsCompatible(existingMethod, inheritedMethod) && !canOverridingMethodDifferInErasure(existingMethod, inheritedMethod)) {
              if (inheritedMethod.isDefault())
                if (inheritedMethod.isAbstract()) {
                  checkPackagePrivateAbstractMethod(inheritedMethod);
                } else if (existingMethod.declaringClass.fPackage != inheritedMethod.declaringClass.fPackage && 
                  this.type.fPackage == inheritedMethod.declaringClass.fPackage && !areReturnTypesCompatible(inheritedMethod, existingMethod)) {
                  continue;
                }  
              if (TypeBinding.notEquals(inheritedMethod.returnType.erasure(), existingMethod.returnType.erasure())) {
                if (areReturnTypesCompatible(existingMethod, inheritedMethod)) {
                  addBridgeMethodCandidate(inheritedMethod);
                  break label117;
                } 
                break label117;
              } 
              break label117;
            } 
            continue;
          }  
        if (!inheritedMethod.isDefault() || inheritedMethod.declaringClass.fPackage == this.type.fPackage) {
          if (existingMethods == null) {
            existingMethods = new MethodBinding[] { inheritedMethod };
          } else {
            int length = existingMethods.length;
            System.arraycopy(existingMethods, 0, existingMethods = new MethodBinding[length + 1], 0, length);
            existingMethods[length] = inheritedMethod;
          } 
          this.inheritedMethods.put(inheritedMethod.selector, existingMethods);
          continue;
        } 
        MethodBinding[] nonVisible = (MethodBinding[])nonVisibleDefaultMethods.get(inheritedMethod.selector);
        if (nonVisible != null && inheritedMethod.isAbstract())
          for (int j = 0, l = nonVisible.length; j < l; j++) {
            if (areMethodsCompatible(nonVisible[j], inheritedMethod))
              break label117; 
          }  
        if (nonVisible == null) {
          nonVisible = new MethodBinding[] { inheritedMethod };
        } else {
          int length = nonVisible.length;
          System.arraycopy(nonVisible, 0, nonVisible = new MethodBinding[length + 1], 0, length);
          nonVisible[length] = inheritedMethod;
        } 
        nonVisibleDefaultMethods.put(inheritedMethod.selector, nonVisible);
        if (inheritedMethod.isAbstract() && !this.type.isAbstract())
          problemReporter().abstractMethodCannotBeOverridden(this.type, inheritedMethod); 
        MethodBinding[] current = (MethodBinding[])this.currentMethods.get(inheritedMethod.selector);
        if (current != null && !inheritedMethod.isStatic()) {
          int j;
          int length;
          label117: for (j = 0, length = current.length; j < length; j++) {
            if (!current[j].isStatic() && areMethodsCompatible(current[j], inheritedMethod)) {
              problemReporter().overridesPackageDefaultMethod(current[j], inheritedMethod);
              break;
            } 
          } 
        } 
      } 
      superType = superType.superclass();
    } 
    List<ReferenceBinding> superIfcList = new ArrayList();
    HashSet seenTypes = new HashSet();
    collectAllDistinctSuperInterfaces(superInterfaces, seenTypes, superIfcList);
    ReferenceBinding currentSuper = superclass;
    while (currentSuper != null && currentSuper.id != 1) {
      collectAllDistinctSuperInterfaces(currentSuper.superInterfaces(), seenTypes, superIfcList);
      currentSuper = currentSuper.superclass();
    } 
    if (superIfcList.size() == 0)
      return; 
    if (superIfcList.size() == 1) {
      superInterfaces = new ReferenceBinding[] { superIfcList.get(0) };
    } else {
      superInterfaces = superIfcList.<ReferenceBinding>toArray(new ReferenceBinding[superIfcList.size()]);
      superInterfaces = Sorting.sortTypes(superInterfaces);
    } 
    SimpleSet skip = findSuperinterfaceCollisions(superclass, superInterfaces);
    int len = superInterfaces.length;
    for (int i = len - 1; i >= 0; i--) {
      superType = superInterfaces[i];
      if (superType.isValidBinding() && (
        skip == null || !skip.includes(superType))) {
        MethodBinding[] methods = superType.unResolvedMethods();
        int m;
        label124: for (m = methods.length; --m >= 0; ) {
          MethodBinding inheritedMethod = methods[m];
          if (inheritedMethod.isStatic())
            continue; 
          MethodBinding[] existingMethods = (MethodBinding[])this.inheritedMethods.get(inheritedMethod.selector);
          if (existingMethods == null) {
            existingMethods = new MethodBinding[] { inheritedMethod };
          } else {
            int length = existingMethods.length;
            for (int e = 0; e < length; e++) {
              if (isInterfaceMethodImplemented(inheritedMethod, existingMethods[e], superType)) {
                if (TypeBinding.notEquals(inheritedMethod.returnType.erasure(), (existingMethods[e]).returnType.erasure()))
                  addBridgeMethodCandidate(inheritedMethod); 
                if (!canOverridingMethodDifferInErasure(existingMethods[e], inheritedMethod))
                  continue label124; 
              } 
            } 
            System.arraycopy(existingMethods, 0, existingMethods = new MethodBinding[length + 1], 0, length);
            existingMethods[length] = inheritedMethod;
          } 
          this.inheritedMethods.put(inheritedMethod.selector, existingMethods);
        } 
      } 
    } 
  }
  
  void collectAllDistinctSuperInterfaces(ReferenceBinding[] superInterfaces, Set<ReferenceBinding> seen, List<ReferenceBinding> result) {
    int length = superInterfaces.length;
    for (int i = 0; i < length; i++) {
      ReferenceBinding superInterface = superInterfaces[i];
      if (seen.add(superInterface)) {
        result.add(superInterface);
        collectAllDistinctSuperInterfaces(superInterface.superInterfaces(), seen, result);
      } 
    } 
  }
  
  protected boolean canOverridingMethodDifferInErasure(MethodBinding overridingMethod, MethodBinding inheritedMethod) {
    return false;
  }
  
  void computeMethods() {
    MethodBinding[] methods = this.type.methods();
    int size = methods.length;
    this.currentMethods = new HashtableOfObject((size == 0) ? 1 : size);
    for (int m = size; --m >= 0; ) {
      MethodBinding method = methods[m];
      if (!method.isConstructor() && !method.isDefaultAbstract()) {
        MethodBinding[] existingMethods = (MethodBinding[])this.currentMethods.get(method.selector);
        if (existingMethods == null) {
          existingMethods = new MethodBinding[1];
        } else {
          System.arraycopy(existingMethods, 0, 
              existingMethods = new MethodBinding[existingMethods.length + 1], 0, existingMethods.length - 1);
        } 
        existingMethods[existingMethods.length - 1] = method;
        this.currentMethods.put(method.selector, existingMethods);
      } 
    } 
  }
  
  MethodBinding computeSubstituteMethod(MethodBinding inheritedMethod, MethodBinding currentMethod) {
    return computeSubstituteMethod(inheritedMethod, currentMethod, this.environment);
  }
  
  public static MethodBinding computeSubstituteMethod(MethodBinding inheritedMethod, MethodBinding currentMethod, LookupEnvironment environment) {
    if (inheritedMethod == null)
      return null; 
    if (currentMethod.parameters.length != inheritedMethod.parameters.length)
      return null; 
    if (currentMethod.declaringClass instanceof BinaryTypeBinding)
      ((BinaryTypeBinding)currentMethod.declaringClass).resolveTypesFor(currentMethod); 
    if (inheritedMethod.declaringClass instanceof BinaryTypeBinding)
      ((BinaryTypeBinding)inheritedMethod.declaringClass).resolveTypesFor(inheritedMethod); 
    TypeVariableBinding[] inheritedTypeVariables = inheritedMethod.typeVariables;
    int inheritedLength = inheritedTypeVariables.length;
    if (inheritedLength == 0)
      return inheritedMethod; 
    TypeVariableBinding[] typeVariables = currentMethod.typeVariables;
    int length = typeVariables.length;
    if (length == 0)
      return inheritedMethod.asRawMethod(environment); 
    if (length != inheritedLength)
      return inheritedMethod; 
    TypeBinding[] arguments = new TypeBinding[length];
    System.arraycopy(typeVariables, 0, arguments, 0, length);
    ParameterizedGenericMethodBinding substitute = 
      environment.createParameterizedGenericMethod(inheritedMethod, arguments);
    for (int i = 0; i < inheritedLength; i++) {
      TypeVariableBinding inheritedTypeVariable = inheritedTypeVariables[i];
      TypeVariableBinding typeVariable = (TypeVariableBinding)arguments[i];
      if (TypeBinding.equalsEquals(typeVariable.firstBound, inheritedTypeVariable.firstBound)) {
        if (typeVariable.firstBound == null)
          continue; 
      } else if (typeVariable.firstBound != null && inheritedTypeVariable.firstBound != null && 
        typeVariable.firstBound.isClass() != inheritedTypeVariable.firstBound.isClass()) {
        return inheritedMethod;
      } 
      if (TypeBinding.notEquals(Scope.substitute((Substitution)substitute, (TypeBinding)inheritedTypeVariable.superclass), (TypeBinding)typeVariable.superclass))
        return inheritedMethod; 
      int interfaceLength = inheritedTypeVariable.superInterfaces.length;
      ReferenceBinding[] interfaces = typeVariable.superInterfaces;
      if (interfaceLength != interfaces.length)
        return inheritedMethod; 
      for (int j = 0; j < interfaceLength; j++) {
        TypeBinding superType = Scope.substitute((Substitution)substitute, (TypeBinding)inheritedTypeVariable.superInterfaces[j]);
        int k = 0;
        while (true) {
          if (k >= interfaceLength)
            return inheritedMethod; 
          if (TypeBinding.equalsEquals(superType, (TypeBinding)interfaces[k]))
            break; 
          k++;
        } 
      } 
      continue;
    } 
    return (MethodBinding)substitute;
  }
  
  static boolean couldMethodOverride(MethodBinding method, MethodBinding inheritedMethod) {
    if (!CharOperation.equals(method.selector, inheritedMethod.selector))
      return false; 
    if (method == inheritedMethod || method.isStatic() || inheritedMethod.isStatic())
      return false; 
    if (inheritedMethod.isPrivate())
      return false; 
    if (inheritedMethod.isDefault() && method.declaringClass.getPackage() != inheritedMethod.declaringClass.getPackage())
      return false; 
    if (!method.isPublic()) {
      if (inheritedMethod.isPublic())
        return false; 
      if (inheritedMethod.isProtected() && !method.isProtected())
        return false; 
    } 
    return true;
  }
  
  public boolean doesMethodOverride(MethodBinding method, MethodBinding inheritedMethod) {
    return doesMethodOverride(method, inheritedMethod, this.environment);
  }
  
  public static boolean doesMethodOverride(MethodBinding method, MethodBinding inheritedMethod, LookupEnvironment environment) {
    return (couldMethodOverride(method, inheritedMethod) && areMethodsCompatible(method, inheritedMethod, environment));
  }
  
  SimpleSet findSuperinterfaceCollisions(ReferenceBinding superclass, ReferenceBinding[] superInterfaces) {
    return null;
  }
  
  MethodBinding findBestInheritedAbstractOrDefaultMethod(MethodBinding[] methods, int length) {
    for (int i = 0; i < length; i++) {
      MethodBinding method = methods[i];
      if (method.isAbstract() || method.isDefaultMethod()) {
        int j = 0;
        while (true) {
          if (j >= length)
            return method; 
          if (i != j && !checkInheritedReturnTypes(method, methods[j])) {
            if (this.type.isInterface() && (methods[j]).declaringClass.id == 1)
              return method; 
            break;
          } 
          j++;
        } 
      } 
    } 
    return null;
  }
  
  int[] findOverriddenInheritedMethods(MethodBinding[] methods, int length) {
    int[] toSkip = null;
    int i = 0;
    ReferenceBinding declaringClass = (methods[i]).declaringClass;
    if (!declaringClass.isInterface()) {
      ReferenceBinding declaringClass2 = (methods[++i]).declaringClass;
      while (TypeBinding.equalsEquals((TypeBinding)declaringClass, (TypeBinding)declaringClass2)) {
        if (++i == length)
          return null; 
        declaringClass2 = (methods[i]).declaringClass;
      } 
      if (!declaringClass2.isInterface()) {
        if (declaringClass.fPackage != declaringClass2.fPackage && methods[i].isDefault())
          return null; 
        toSkip = new int[length];
        do {
          toSkip[i] = -1;
          if (++i == length)
            return toSkip; 
          declaringClass2 = (methods[i]).declaringClass;
        } while (!declaringClass2.isInterface());
      } 
    } 
    for (; i < length; i++) {
      if (toSkip == null || toSkip[i] != -1) {
        declaringClass = (methods[i]).declaringClass;
        for (int j = i + 1; j < length; j++) {
          if (toSkip == null || toSkip[j] != -1) {
            ReferenceBinding declaringClass2 = (methods[j]).declaringClass;
            if (!TypeBinding.equalsEquals((TypeBinding)declaringClass, (TypeBinding)declaringClass2))
              if (declaringClass.implementsInterface(declaringClass2, true)) {
                if (toSkip == null)
                  toSkip = new int[length]; 
                toSkip[j] = -1;
              } else if (declaringClass2.implementsInterface(declaringClass, true)) {
                if (toSkip == null)
                  toSkip = new int[length]; 
                toSkip[i] = -1;
                break;
              }  
          } 
        } 
      } 
    } 
    return toSkip;
  }
  
  boolean isAsVisible(MethodBinding newMethod, MethodBinding inheritedMethod) {
    if (inheritedMethod.modifiers == newMethod.modifiers)
      return true; 
    if (newMethod.isPublic())
      return true; 
    if (inheritedMethod.isPublic())
      return false; 
    if (newMethod.isProtected())
      return true; 
    if (inheritedMethod.isProtected())
      return false; 
    return !newMethod.isPrivate();
  }
  
  boolean isInterfaceMethodImplemented(MethodBinding inheritedMethod, MethodBinding existingMethod, ReferenceBinding superType) {
    return (areParametersEqual(existingMethod, inheritedMethod) && existingMethod.declaringClass.implementsInterface(superType, true));
  }
  
  public boolean isMethodSubsignature(MethodBinding method, MethodBinding inheritedMethod) {
    return (CharOperation.equals(method.selector, inheritedMethod.selector) && 
      isParameterSubsignature(method, inheritedMethod));
  }
  
  boolean isParameterSubsignature(MethodBinding method, MethodBinding inheritedMethod) {
    return isParameterSubsignature(method, inheritedMethod, this.environment);
  }
  
  static boolean isParameterSubsignature(MethodBinding method, MethodBinding inheritedMethod, LookupEnvironment environment) {
    MethodBinding substitute = computeSubstituteMethod(inheritedMethod, method, environment);
    return (substitute != null && isSubstituteParameterSubsignature(method, substitute, environment));
  }
  
  boolean isSubstituteParameterSubsignature(MethodBinding method, MethodBinding substituteMethod) {
    return isSubstituteParameterSubsignature(method, substituteMethod, this.environment);
  }
  
  public static boolean isSubstituteParameterSubsignature(MethodBinding method, MethodBinding substituteMethod, LookupEnvironment environment) {
    if (!areParametersEqual(method, substituteMethod)) {
      if (substituteMethod.hasSubstitutedParameters() && method.areParameterErasuresEqual(substituteMethod))
        return (method.typeVariables == Binding.NO_TYPE_VARIABLES && !hasGenericParameter(method)); 
      if (method.declaringClass.isRawType() && substituteMethod.declaringClass.isRawType() && 
        method.hasSubstitutedParameters() && substituteMethod.hasSubstitutedParameters())
        return areMethodsCompatible(method, substituteMethod, environment); 
      return false;
    } 
    if (substituteMethod instanceof ParameterizedGenericMethodBinding) {
      if (method.typeVariables != Binding.NO_TYPE_VARIABLES)
        return !((ParameterizedGenericMethodBinding)substituteMethod).isRaw; 
      return !hasGenericParameter(method);
    } 
    return (method.typeVariables == Binding.NO_TYPE_VARIABLES);
  }
  
  static boolean hasGenericParameter(MethodBinding method) {
    if (method.genericSignature() == null)
      return false; 
    TypeBinding[] params = method.parameters;
    for (int i = 0, l = params.length; i < l; i++) {
      TypeBinding param = params[i].leafComponentType();
      if (param instanceof ReferenceBinding) {
        int modifiers = ((ReferenceBinding)param).modifiers;
        if ((modifiers & 0x40000000) != 0)
          return true; 
      } 
    } 
    return false;
  }
  
  boolean isSameClassOrSubclassOf(ReferenceBinding testClass, ReferenceBinding superclass) {
    while (true) {
      if (TypeBinding.equalsEquals((TypeBinding)testClass, (TypeBinding)superclass))
        return true; 
      if ((testClass = testClass.superclass()) == null)
        return false; 
    } 
  }
  
  boolean mustImplementAbstractMethod(ReferenceBinding declaringClass) {
    if (!mustImplementAbstractMethods())
      return false; 
    ReferenceBinding superclass = this.type.superclass();
    if (declaringClass.isClass()) {
      while (superclass.isAbstract() && TypeBinding.notEquals((TypeBinding)superclass, (TypeBinding)declaringClass))
        superclass = superclass.superclass(); 
    } else {
      if (this.type.implementsInterface(declaringClass, false) && 
        !superclass.implementsInterface(declaringClass, true))
        return true; 
      while (superclass.isAbstract() && !superclass.implementsInterface(declaringClass, false))
        superclass = superclass.superclass(); 
    } 
    return superclass.isAbstract();
  }
  
  boolean mustImplementAbstractMethods() {
    return (!this.type.isInterface() && !this.type.isAbstract());
  }
  
  ProblemReporter problemReporter() {
    return this.type.scope.problemReporter();
  }
  
  ProblemReporter problemReporter(MethodBinding currentMethod) {
    ProblemReporter reporter = problemReporter();
    if (TypeBinding.equalsEquals((TypeBinding)currentMethod.declaringClass, (TypeBinding)this.type) && currentMethod.sourceMethod() != null)
      reporter.referenceContext = (ReferenceContext)currentMethod.sourceMethod(); 
    return reporter;
  }
  
  boolean reportIncompatibleReturnTypeError(MethodBinding currentMethod, MethodBinding inheritedMethod) {
    problemReporter(currentMethod).incompatibleReturnType(currentMethod, inheritedMethod);
    return true;
  }
  
  ReferenceBinding[] resolvedExceptionTypesFor(MethodBinding method) {
    ReferenceBinding[] exceptions = method.thrownExceptions;
    if ((method.modifiers & 0x2000000) == 0)
      return exceptions; 
    if (!(method.declaringClass instanceof BinaryTypeBinding))
      return Binding.NO_EXCEPTIONS; 
    for (int i = exceptions.length; --i >= 0;)
      exceptions[i] = (ReferenceBinding)BinaryTypeBinding.resolveType((TypeBinding)exceptions[i], this.environment, true); 
    return exceptions;
  }
  
  void verify() {
    computeMethods();
    computeInheritedMethods();
    checkMethods();
    if (this.type.isClass())
      checkForMissingHashCodeMethod(); 
  }
  
  void verify(SourceTypeBinding someType) {
    if (this.type == null) {
      try {
        this.type = someType;
        verify();
      } finally {
        this.type = null;
      } 
    } else {
      this.environment.newMethodVerifier().verify(someType);
    } 
  }
  
  public String toString() {
    StringBuffer buffer = new StringBuffer(10);
    buffer.append("MethodVerifier for type: ");
    buffer.append(this.type.readableName());
    buffer.append('\n');
    buffer.append("\t-inherited methods: ");
    buffer.append(this.inheritedMethods);
    return buffer.toString();
  }
}
